
/* private include file for db package */

#ifndef __P_DB__
#define __P_DB__

#include <unistd.h>
#include <pthread.h>
#include <stdio.h>
#include <sys/stat.h>
#include "db.h"
#include "p_sort.h"

#define NFS -2
#define LOCAL -1
#define NOLOCKS 0

#define MAX_DATABASES            1000

#define MAX_DESCRIPTOR_SIZE 	2048

extern char    *db_nullstr;
#define Zfree(address) if(address != db_nullstr) {free(address); address = NULL;}

/* Following are the structs which define the data structures used by the db
 * package. */

typedef struct Attribute
{				       /* description of an attribute */
    char           *name;	       /* name of attribute */
    long            id;		       /* id # of attribute */
    long            size;	       /* number of characters */
    int            type;	       /* type  */
    long            added;	       /* flag for "added" (not in original schema) attribute */
    char           *null;	       /* null value string */
    char           *fmt;	       /* printf style format for writing
				        * into table */
    char           *range;	       /* valid range of field */
    char           *units;	       /* units of field */
    char           *desc;	       /* short description */
    char           *detail;	       /* long description */
    char           *base_table;	       /* for link attribute, specifies
				        * defining table */
    Tbl            *relations;	       /* list of relations using attribute */
}               Attribute;

typedef struct Relation
{				       /* description of relation */
    char           *name;	       /* name of relation */
    long            id;		       /* id # for relation in schema */
    char            fldsep;	       /* separator between fields */
    char            recsep;	       /* separator between records */
    long            size;	       /* size of record in relation */
    long            added;	       /* flag for "added" (not in original schema) relation */

    char           *null;	       /* null record */
    char 	   *like; 	       /* name of "like" relation, if it exists */

    long	    transient;	       /* flag = 1 if relation is not mapped to disc */
    long            view;	       /* flag = 1 if relation is a view */
    long            direct;	       /* flag = 1 if relation has "direct"
				        * field */

    long            fields_count;      /* count of the number of fields */
    long            all_fields_count;  /* count of fields, including indirect
				        * relation references */
    Arr            *fields_root;       /* array of Fields indexed by field
				        * names, including relation names */
    Tbl            *fields_tbl;	       /* table of fields */
    Tbl            *fieldnames_tbl;    /* table of field names (no relation
				        * names) */
    Tbl		   *fieldlist;	       /* table of fields from schema --
    					  may be just list of relations */

    long            rels_count ;       /* count of (other) relations */
    Tbl 	   *relnames_tbl;      /* table of relation names which form
					* view (not including this relation) */
    Tbl            *relref_tbl;        /* table of RelRef's which form table
					* (including this relation) */

    Tbl            *primary_keys;      /* table of primary key (names) */
    Tbl            *alternate_keys;    /* table of alternate key (names) */
    Tbl            *foreign_keys;      /* table of foreign key (names) */
    Tbl		   *other_key;        /* table of alternate keys */

    char           *unique_name;       /* name of unique integer id key */

    char           *desc;	       /* short description */
    char           *detail;	       /* long description */
}               Relation;

#define dbREF_DIRECT	0	       /* field in this table */
#define dbREF_INDIRECT	1	       /* field in another table */
#define dbREF_TABLE 	2	       /* another table */
#define dbREF_SELF	3	       /* this table */

typedef struct RelRef
{
    Relation       *rel;	       /* reference to relation */
    long            id_offset;	       /* add to field id to get id of field
				        * in this table */
    long            reftype;	       /* dbREF_* type */
    long	    disp;	       /* displacement of dbptr in record */
}               RelRef;

typedef struct Schema
{
    char           *name;	       /* Schema name */
    char           *path;	       /* schema line from descriptor file */
    char	   *ext;	       /* directory for extensions */

    char           *desc;	       /* short description */
    char           *detail;	       /* long description */
    long            id;		       /* Schema id */

    char	   *timedate ; 		/* timedate field name */
    long            // rels_count,
                    attrs_count,
                    dbs_count;
    Arr            *attrs_root;	       /* root of binary tree of attributes */
    Tbl            *attrs_tbl;	       /* table of attribute names */
    Arr            *links_arr;	       /* table of link attributes */

    Arr            *rels_root;	       /* root of binary tree of relations */
    Tbl            *rels_tbl;	       /* table of relations */
    Tbl            *relnames_tbl;      /* table of loaded relation names */
    Tbl		   *alltables ;        /* including extensions */
    int		    added ; 	       /* set to 1 in open_schema */
}               Schema;

typedef struct Field
{				       /* description of single field in a
				        * particular table */
    long            id;		       /* id # for field in table */
    long            disp;	       /* position of field in record */
    long            relid;	       /* id of base relation in which field
				        * is found */
    long            tableref;	       /* flag = 1 if field is table */
    union
      {
	Attribute      *attr;	       /* information about attribute */
	Relation       *rel;	       /* or relation */
      }               ref;
}               Field;

typedef struct Table
{				       /* corresponding to actual file */
    char           *name;	       /* table name */
    long            id;		       /* table id */
    Relation       *rel;	       /* pointer to relation */
    char           *creation;	       /* description of how to create table */
    Arr            *index_arr;	       /* indexes on this table */
    long            count;	       /* count of records in table */

    char           *path;	       /* name of physical file */
    char           *dir;	       /* directory of file -- for file
				        * references */

    int            fd;		       /* open file descriptor */

    size_t          size;	       /* length of mmap'ed data */
    size_t          maxsize;	       /* malloc'd size for a view only */
    char           *data;	       /* pointer to mmap'ed data */

    char	   *lockfile ; 	       /* path to lockfile */
    int	            lockfd ;	       /* open file descriptor for lockfile */

    long            writeable;	       /* flag = true if file is writeable */

    Dbindex        *dbindex;
    long           *tr_database;       /* translate table for database id */
    long           *tr_table;	       /* translate table for table id */
    char           *scratch;	       /* scratch record */
}               Table;

typedef struct Database
{				       /* corresponding to set of files */
    char           *name;	       /* name of database */
    long            id;		       /* database id */
    char	   *schema_list ;      /* schema name from database descriptor */
    Schema         *schema;	       /* schema for a database */
    char           *dbpath;	       /* path for looking up tables */

    char           *path;	       /* name of physical file */
    char           *dir;	       /* directory of file -- for file
				        * references */

    char           *openperm;	       /* permissions for opening tables */
    int             locks ;		/* flag != 0 when database is to be
					    locked before writing */
    char	   *idserver ; 		/* when idserver != 0, specifies 
					    host:port to get id's from */

    char           *desc;
    char           *detail;

    Arr	           *lastids ; 		/* array of lastid's indexed by id name */
    Tbl            *tables_tbl;	       /* array of tables in database */
    Arr		   *files ; 
}               Database;

/* Globals */

extern char    *Default_schema;
extern char    *Default_dbpath;
extern int     Db_dbmatches_update_index ; 

extern pthread_mutex_t  DB_global_mutex ;
extern Tbl     *Schemas_tbl;
extern Arr     *Databases_root;
extern Database *Database_List[MAX_DATABASES] ;
extern int       Database_Id ; 
extern Attribute p_dbptr ;
extern Tbl      _db_nulltbl ; 
extern long     Db_Count ;

#define dbDATABASE		-401
#define dbTABLE			-403
#define dbFIELD			-404
#define dbRECORD		-405

#ifdef lint
extern int      ZERO;

#else
#define ZERO 0
#endif


#ifdef  __cplusplus
extern "C" {
#endif

extern int compile_schema ( Schema *schema, int fd, char *s, char *database_name, char *schemafile ) ;

extern int dbadd_attribute ( Schema *schema, Attribute *attribute );
extern int dbadd_like_relation ( Schema *schema, Relation *relation, char *oldrel_name );
extern int dbadd_relation ( Schema *schema, Relation *relation );
extern void add2index ( char *index_name, void *dbindex, void *dbp );
extern int dbadd2index ( Dbptr db, Dbindex *dbindex );
extern int dbc2d ( Dbvalue *value, char *location, int length, char *fmt, int type );
extern int dbd2c ( char *location, int length, Dbvalue *value, int type );
extern int dbc2p ( Dbvalue *value, char *location, int length, char *fmt, int type );
extern int dbp2c ( char *location, int length, Dbvalue *value, int type );
extern int dbcmp ( char *key1, char *key2, void *dbindex );
extern int dbcmp_debug_search ( char *key1, char *key2, void *dbindex );
extern int dbcmp_envelope ( char *key1, char *key2, void *dbindex );
extern int dbcmp_search ( char *key1, char *key2, void *dbindex );
extern int dbcmp_sort ( char *key1, char *key2, void *dbindex );
extern int dbcmp_psort (void *vindex, char *key1, char *key2 ) ;
extern Dbindex * dbcreate_index ( Dbptr db, Tbl *pattern, char *name );
extern Dbindex * dbcreate_index2 ( Dbptr db, Tbl *pattern, char *name, long *ret );
extern int dbfield_type ( Dbptr db, char *fieldname );
extern int dbfix_envelope ( Dbindex *dbindex );
extern void dbfree_jig ( Dbjig *jig );
extern void dbfree_index ( Dbindex *dbindex );
extern void dbfree_attribute ( Attribute *attr );
extern void dbfree_field ( Field *field );
extern void dbfree_relation ( Relation *rel );
extern int dbfree_table ( Table * table );
extern int dbfree_database ( Database *database );
extern void dbfree_schema ( Schema *schema );
extern int dbget_index ( Dbptr db, Tbl *pattern, Dbindex **index );
extern int dbget_index2 ( Dbptr db, Tbl *pattern, Dbindex **index );
extern char * jointbl ( Tbl *tbl, char *separator );
extern Dbptr dbmkgroup_relation ( Dbptr db, Tbl *groupfields, char *name );
extern int dbindex_name ( Dbptr db, Tbl *pattern, char *s );
extern int dbmake_jig ( Dbptr db, Tbl *pattern, Dbjig **jigp );
extern void dbshow_key ( Dbjig * jig, char *key );
extern void dbshow_null (Dbptr db) ;
extern void dbmake_key ( Dbptr db, Dbjig *jig, char *key );
extern long dbmake_key2 ( Dbptr db, Dbjig *jig, char *key );
extern int dbprint ( FILE *file, Dbptr db, int *ids, int n );
extern int dbprt_key ( char *key, Dbjig *jig );
extern int dbprt_index ( Dbindex *dbindex );
extern Dbindex *dbread_index ( Dbptr db, Table *table, char *filename );
extern int dbsize ( Dbptr db );
extern long dbrecord ( Dbindex *dbindex, long j );
extern int dbresolve ( Dbptr db, Database **database, Relation **relation, Table **table, Field **field, char **location, long *length );
extern int dbsave_index ( Dbptr db, Dbindex *dbindex );
extern int dbsearch ( Dbptr db, Dbjig *jig, Dbindex *dbindex, int unique, long *ns, long *ne );
extern char * dbtmpname ( Dbptr db, char *name );
extern int dbupdate_index ( Dbindex *dbindex );
extern Dbptr dbtbl2view ( Dbptr db, char *name, Tbl *tables );
extern char * equiv_field ( char *field );
extern Expression *ex_to_type ( Expression *node, int type );
extern int eqfunct ( char *s );

extern void exshow_immediate ( Exvalue * value, int type );
extern void exshow_type ( int type );
extern void exshow_value ( Exvalue * value, int kind, int type );
extern void exshow_expression ( Expression * tree, int flag );

extern int find_path ( char *databasename, char *tablename, char *dbpath, char *tblpath );
extern int find_schemafile ( char *filename, char *schemaname, char *databasename );
extern int map_table ( Database *database, Relation *relation, int createflag );
extern Table * new_table ( void );
extern Database * new_database ( void );
extern Schema * new_schema ( void );
extern Attribute * new_attribute ( void );
extern Relation * new_relation ( void );
extern Field * new_field ( void );
extern void new_scratch ( Dbptr db, Table *table );
extern int new_null ( Dbptr db );

extern Schema * open_schema ( char *schema_list, char *databasename );
extern int add_extension ( Dbptr db, Schema *schema, char *table );

extern Expression *exnew ( int type, int kind, int integer );
extern int _dblock( Database *database, Table *table );
extern int _dbunlock( Database *database, Table *table );

extern void show_relation ( Dbptr db, Relation * rel );
extern void show_schema ( Dbptr db );
extern void show_field ( Dbptr db, Field * field );
extern void dbshow_index ( char *name, void *dbindex, void *pvt );
extern void show_table ( Dbptr db, Table * table );
extern int show_view ( Dbptr db, FILE * file );
extern int dbshow ( Dbptr db );
extern void dbshow_record ( Dbptr db );
extern void dbshow_database (Dbptr db) ;

extern int _dbex_eval ( Dbptr db, Expression * ex, int setflag, void *vresult );

extern char *dblockname ( char *path );
extern Arr * dbtablefiles ( Dbptr db );

extern void dbkeys2stk ( void **stk, Tbl **keys1p, Tbl **keys2p );
extern int dbadd2creation ( Dbptr parent, Dbptr db, char *astep );

extern int dbfix_nullrecords (Dbptr db) ;
extern char *new_anonymous_name() ;
extern long dbnext_free_relation_id(Schema *schema)  ;
extern long dbnext_free_relation_id_from_db ( Dbptr db ) ;

#ifdef  __cplusplus
}
#endif

#endif
